import os
import functools
import operator
import gzip
import struct
import array
import numpy as np
from urllib.request import urlretrieve
import sys

#from: https://github.com/datapythonista/mnist
DATASET_DIRECTORY = 'data/'
URL = 'http://yann.lecun.com/exdb/mnist/'

class IdxDecodeError(ValueError):
    """Raised when an invalid idx file is parsed."""
    pass

def parse_idx(fd):
    DATA_TYPES = {0x08: 'B',  # unsigned byte
                  0x09: 'b',  # signed byte
                  0x0b: 'h',  # short (2 bytes)
                  0x0c: 'i',  # int (4 bytes)
                  0x0d: 'f',  # float (4 bytes)
                  0x0e: 'd'}  # double (8 bytes)

    header = fd.read(4)
    if len(header) != 4:
        raise IdxDecodeError('Invalid IDX file, file empty or does not contain a full header.')

    zeros, data_type, num_dimensions = struct.unpack('>HBB', header)

    if zeros != 0:
        raise IdxDecodeError('Invalid IDX file, file must start with two zero bytes. '
                             'Found 0x%02x' % zeros)

    try:
        data_type = DATA_TYPES[data_type]
    except KeyError:
        raise IdxDecodeError('Unknown data type 0x%02x in IDX file' % data_type)

    dimension_sizes = struct.unpack('>' + 'I' * num_dimensions,
                                    fd.read(4 * num_dimensions))

    data = array.array(data_type, fd.read())
    data.byteswap()  # looks like array.array reads data as little endian

    expected_items = functools.reduce(operator.mul, dimension_sizes)
    if len(data) != expected_items:
        raise IdxDecodeError('IDX file has wrong number of items. '
                             'Expected: %d. Found: %d' % (expected_items, len(data)))

    return np.array(data).reshape(dimension_sizes)

def print_download_progress(count, block_size, total_size):
    pct_complete = int(count * block_size * 100 / total_size)
    pct_complete = min(pct_complete, 100)
    msg = "\r- Download progress: %d" % (pct_complete) + "%"
    sys.stdout.write(msg)
    sys.stdout.flush()


def download_and_parse_mnist_file(fname, target_dir=None, force=False):
    if not os.path.exists(DATASET_DIRECTORY):
        os.makedirs(DATASET_DIRECTORY)
    if not os.path.exists(DATASET_DIRECTORY+fname):
        print('Downloading '+fname)
        file_path = os.path.join(DATASET_DIRECTORY, fname)
        url = URL + fname
        file_path, _ = urlretrieve(url=url, filename=file_path, reporthook=print_download_progress)
        print("\nDownload finished.")

    fname = 'data/' + fname
    fopen = gzip.open if os.path.splitext(fname)[1] == '.gz' else open
    with fopen(fname, 'rb') as fd:
        return parse_idx(fd)


def train_images():
    return download_and_parse_mnist_file('train-images-idx3-ubyte.gz')


def test_images():
    return download_and_parse_mnist_file('t10k-images-idx3-ubyte.gz')


def train_labels():
    return download_and_parse_mnist_file('train-labels-idx1-ubyte.gz')


def test_labels():
    return download_and_parse_mnist_file('t10k-labels-idx1-ubyte.gz')


from PyEZNet.Diff_Func import Diff_Func
import numpy as np
from math import sqrt
from itertools import product
from PyEZNet.utils import ZeroPadding


class Layer (Diff_Func):
    '''
    Abstract Class to represent Layers so we added 
        1- Weights dic
        2- Weights update dic
        3- Init weights Function
        4- Update_Weights Function    
    '''
    def __init__(self, *args, **kwargs):
        # Call Diff_Func init function
        super().__init__(*args, **kwargs)
        # Empty Weights and updates
        self.weights = {}
        self.weights_Update = {}
        
     # Function to initalize weights   
    def init_Weights(self , *args, **kwargs):
         pass
     
    def Update_Weights(self , learningRate):
         
        for weight_key , weight in self.weights.items():
            self.weights[weight_key] = self.weights[weight_key] - learningRate * self.weights_Update[weight_key]

class MaxPool2D(Diff_Func):
    ###############################################################################
    #
    # Class: MaxPool2D
    #
    # File Name: layer.py
    #
    # Description: class for the maximum pooling layer.
    # 
    # Author: Ahmed Gamal
    ################################################################################
    def __init__(self, kernel_size=(2, 2)):
        super().__init__()
        self.kernel_size = (kernel_size, kernel_size) if isinstance(kernel_size, int) else kernel_size
        #isinstance ->  check if the kernel size is int number 
    def __call__(self, X):
        # in contrary to other Function subclasses, MaxPool2D does not need to call
        # .local_grad() after forward pass because the gradient is calculated during it
        return self.forward(X)
    def forward(self, X):
        N, C, H, W = X.shape
        KH, KW = self.kernel_size

        grad = np.zeros_like(X)
        Y = np.zeros((N, C, H//KH, W//KW)) #h/kh -> hight of the new matrix ,, w/kw -> width of the new matrix

        for h, w in product(range(0, H//KH), range(0, W//KW)):
            h_offset, w_offset = h*KH, w*KW
            rec_field = X[:, :, h_offset:h_offset+KH, w_offset:w_offset+KW] # get the part of matrix that we will find it max 
            Y[:, :, h, w] = np.max(rec_field, axis=(2, 3)) #find the max in the part of matrix
            for kh, kw in product(range(KH), range(KW)):
                grad[:, :, h_offset+kh, w_offset+kw] = (X[:, :, h_offset+kh, w_offset+kw] >= Y[:, :, h, w])

        # storing the gradient
        self.grad['X'] = grad

        return Y
    def backward(self, dY):
        dY = np.repeat(np.repeat(dY, repeats=self.kernel_size[0], axis=2),
                       repeats=self.kernel_size[1], axis=3)
        return self.grad['X']*dY

    def calc_Grad(self, X):
        # small hack: because for MaxPool calculating the gradient is simpler during
        # the forward pass, it is calculated there and this function just returns the
        # grad dictionary
        return self.grad   


class AvePool2D(Diff_Func):
    ###############################################################################
    #
    # Class: AvaragePool2D
    #
    # File Name: layer.py
    #
    # Description: class for the avarage pooling layer.
    # 
    # Author: Dalia Ayman
    ################################################################################
    def __init__(self, kernel_size=(2, 2)):
        super().__init__()
        self.kernel_size = (kernel_size, kernel_size) if isinstance(kernel_size, int) else kernel_size
        #isinstance ->  check if the kernel size is int number 
    def __call__(self, X):
        # in contrary to other Function subclasses, MaxPool2D does not need to call
        # .local_grad() after forward pass because the gradient is calculated during it
        return self.forward(X)
    def forward(self, X):
        N, C, H, W = X.shape
        KH, KW = self.kernel_size

        grad = np.zeros_like(X)
        Y = np.zeros((N, C, H//KH, W//KW)) #h/kh -> hight of the new matrix ,, w/kw -> width of the new matrix

        for h, w in product(range(0, H//KH), range(0, W//KW)):
            h_offset, w_offset = h*KH, w*KW
            rec_field = X[:, :, h_offset:h_offset+KH, w_offset:w_offset+KW] # get the part of matrix that we will find it max 
            Y[:, :, h, w] = np.average(rec_field, axis=(2, 3)) #find the max in the part of matrix
            for kh, kw in product(range(KH), range(KW)):
                grad[:, :, h_offset+kh, w_offset+kw] = (X[:, :, h_offset+kh, w_offset+kw] >= Y[:, :, h, w])

        # storing the gradient
        self.grad['X'] = grad

        return Y
def backward(self, dY):
        dY = np.repeat(np.repeat(dY, repeats=self.kernel_size[0], axis=2),
                       repeats=self.kernel_size[1], axis=3)
        return self.grad['X']*dY

def local_grad(self, X):
        # small hack: because for MaxPool calculating the gradient is simpler during
        # the forward pass, it is calculated there and this function just returns the
        # grad dictionary
        return self.grad   

class Conv2D(Layer):
    """
        Conv2D Takes input img (Channel , Width , Height)  with N imgs -> (N , Ch , H , W)
        and Kernal Size 
        
        Inputs for the layer :
            1-  in_Channels
            2-  out_Channels (Number of Filter in the Conv Layer)
            3-  Padding
            4-  Stride
            5-  Kernal Size (Size of the Filter ex: 3x3 filter)
    """
    def __init__(self , in_Channels , out_Channels , Padding=0 , Stride=1 , Kernal_Size=3):
        # Super Class will creat
        #    ( grad , cache , Weights , Weights_Update ) Empty local dicts
        super().__init__()

        self.in_Channels = in_Channels
        self.Num_Filters = out_Channels
        self.Pad = Padding
        self.Stride = Stride
        self.Filter_Dim = (Kernal_Size,Kernal_Size)

        #Based on the (Filter_Dim, in_Channels, Num_Filters) we init our weights
        self.init_Weights()

    # Generate random values in a normal distribution for ( W ) - zeros for the Bias
    def init_Weights(self):
        scale = 2/sqrt(self.in_Channels * self.Filter_Dim[0] * self.Filter_Dim[1])

        # W size (F , C , W , H)
        self.weights= {'W' : np.random.normal(scale= scale , size= (self.Num_Filters , self.in_Channels ,self.Filter_Dim[0] ,self.Filter_Dim[1] ))
                       ,'b' : np.zeros(shape= (self.Num_Filters , 1)) }
        #self.Weights= {'W' : [[[[-0.2392, -0.0492, -0.0347],
        #                     [-0.3049, -0.1630,  0.1242],
        #                     [-0.2988, -0.3229,  0.1064]]]] ,
        #               'b' : [-0.0967] }


    def forward(self, X):
        '''
            Input :
                - X -> Batch of imgs or single img --   X.shape = (N , Ch , H , W)
            Output :
                - Y -> Output of the Forward Prop with the current W  --  Y.shape = (N , F , H_out , W_out)
                Note :
                    W_out and H_out = floor(( W_in + 2* padding - kernal_size )/ stride + 1)
        '''
        # if we have Padding we first adjust the X
        if self.Pad != 0:
            #Dims always (2,3) H,W
            X = ZeroPadding(X , self.Pad , dims=(2,3))

        #Save input for backProp
        self.cache ={'X' : X}

        N , Ch , H , W = X.shape

        Filter_H , Filter_W = self.Filter_Dim

        # Note : Padding Already Added from the Function Above so we dont need to add it here
        W_out = int(((W - Filter_W )/self.Stride) + 1)
        H_out = int(((H - Filter_H )/self.Stride) + 1)

        Y = np.zeros(shape=(N , self.Num_Filters , H_out ,W_out ))

        #loop over every Img
        for n in range(N):
            #Loop over every filter
            for C_out in range(self.Num_Filters):
                #Loop over H & W
                for h , w in product(range(W_out) , range(H_out)):
                    H_offset , W_offset = h*self.Stride , w*self.Stride
                    # Subset from X for the filter size
                    # Select [ one img (n)  , All the Ch , Offset+ Filter_h , Offset+Filter_W ]
                    Sub_X = X[n , : , H_offset: (H_offset + Filter_H) , W_offset: (W_offset + Filter_W)  ]

                    Y[n , C_out , h , w ] = np.sum(self.weights['W'][C_out] * Sub_X) + self.weights['b'][C_out]
        return Y
    def backward(self , dY):
        '''
            The Goal is to calc :
                1- dW/dL    -> Conv(X , dY/dl)
                2- dB/dL        Conv(X , dY/dl)
                3- dX/dL
        '''
        # Load X from the cache
        X = self.cache['X']
        # dX same shape as X
        dX = np.zeros_like(X)
        N ,CH ,H ,W = X.shape

        Filter_H , Filter_W = self.Filter_Dim

        # Calc dX
        #Loop Over every img
        for n in range(N):
            # Loop over filters
            for C_out in range(self.Num_Filters):
                # Loop over h,w
                for h,w in product(range(dY.shape[2]),range(dY.shape[3])):
                    H_offset , W_offset = h*self.Stride , w*self.Stride
                    #                                                                                          filter index
                    dX[n,: , H_offset:H_offset + Filter_H , W_offset:W_offset + Filter_W ] += self.weights['W'][C_out] * dY[n,C_out,h,w]


        # Calc dW
        dW = np.zeros_like(self.weights['W'])

        # Loop over filters
        for c_w in range(self.Num_Filters):
            for c_i in range(self.in_Channels):
                for h,w in product(range(Filter_H),range(Filter_W)):
                    ######################################################## Msh fahm #####################################################
                    Sub_X = X[: , c_i , h:H-Filter_H+h+1:self.Stride  , w: W-Filter_W+w+1:self.Stride ]
                    dY_rec_field = dY[:, c_w]
                    dW[c_w, c_i, h ,w] = np.sum(Sub_X*dY_rec_field)
                    ##################################################### #####################################################

        # Calc dB
        ######################### Msh Fahmm #############################
        dB = np.sum(dY, axis=(0, 2, 3)).reshape(-1, 1)
        ######################################################
        # caching the global gradients of the parameters
        self.weights_Update['W'] = dW
        self.weights_Update['b'] = dB

        if self.Pad == 0:
            return dX
        return dX[:, :, self.Pad:-self.Pad, self.Pad:-self.Pad]

class FullyConnectedLayer(Layer):

     def __init__(self,input_dim,output_dim):
         super().__init__()      #inheret initial function from our parent class 'Layer'
         self._init_Weights(input_dim,output_dim)


     def _init_Weights(self,input_dim,output_dim):
         scale = 1/ sqrt(input_dim)
         #input_dim= rows , output_dim = columns
         self.weights['W'] = scale *np.random.randn(input_dim,output_dim)
         # 1 = rows , output_dim = columns
         self.weights['b'] = scale *np.random.randn(1,output_dim)
         


     def forward(self,X):

    # forwar function take X 'vector of numpy array' where it's size is (no_of_batch,input_dim)
    # and it's output is Y = (WX+b)
        output=np.dot(X,self.weights['W']) + self.weights['b']
    #cashe it's like a dictionary to save the answer of a frequent qeustions
        self.cache['X']=X
        self.cache['output']=output
        return output

     def backward(self,dY):
    # Our goal with backpropagation is to update each of the weights in the network
    # so that they cause the actual output to be closer the target output using 'Chain Rule'
        dX=dY.dot(self.grad['X'].T)
    # dY is output array of size (rows= no-of_batches, colomns = no_of_out)
    # calculate golbal gradient to be used in backpropagation

        X=self.cache['X']
        dw=self.grad['W'].T.dot(dY)
        
    # calculating the global gradient wrt to weight
        db=np.sum(dY,axis=0,keepdims=True)
    # keepdims=True means that the result will broadcast correctly against the input array.
    # axis=0 means will sum all of the elements of the input array.
    # dY is the elements to sum
        self.weights_Update={'W': dw,'b': db}
        return dX


     def calc_Grad(self,X):

         gradX_local = self.weights['W'] #weight hena gaya mn class function el kber el lsa hyt3mlo implementation
         gradW_local= X
         gradb_local = np.ones_like(self.weights['b'])
         grads = {'X': gradX_local, 'W': gradW_local, 'b': gradb_local}
         return grads



class Flatten(Diff_Func):
    def forward(self, X):
        # Save original size of the input
        self.cache['shape'] = X.shape
        # Get number of imgs 
        n_batch = X.shape[0]
        return X.reshape(n_batch, -1)

    def backward(self, dY):
        return dY.reshape(self.cache['shape'])



from PyEZNet.LOSS_FUNC import Loss
from PyEZNet.Diff_Func import Diff_Func
from PyEZNet.Layer import Layer
import numpy as np
import pickle


class Net:
    #memory optimization for storage
    __slots__ = ['layers','loss_function']

    # layers and loss are parameter
    #while argument is the value sent to the function when it is called
    def __init__(self,layers,loss):
        assert isinstance(loss,Loss) #the loss function must be as instance of nn.losses.Loss
        for layer in layers:
            assert isinstance(layer,Diff_Func)
            #layer must be instance of nn.layers.Layer or nn.layers.Function

        self.layers=layers
        self.loss_function=loss

    def __call__(self, *args, **kwargs):
        '''
        :param args: don't know the no.of arg fe 3mlt * ablha
        :param kwargs: keyword argument w bardo msh 3arf 3ddhum fe 3mltha **
        :return: self.forward
        '''
        return self.forward(*args, **kwargs)

    def forward(self,x):
        '''
        :param x: numpy input array to our net
        :return:  numpy output array
        '''

        for layer in self.layers:
            x= layer(x)
        return x

    def loss(self,x,y):
        '''
        :param x: numpy array output of forward pass
        :param y: numpy array which is true values
        :return: numpy float loss value
        '''
        loss = self.loss_function(x,y)
        return loss


    def backward(self):
        '''
        calculate backward pass for net which must be calculated after calculate forwad pass and loss
        :return: numpy array of shape matching the input during forward pass
        '''
        back=self.loss_function.backward()
        for layer in reversed(self.layers):
            back=layer.backward(back)
        return back


    def weights_update(self,alpha):
        '''
        updates the weights using the gradient which is calculated during backpropagation
        :param alpha: numpy float number
        :return:
        '''
        for layer in self.layers:
            if isinstance(layer,Layer):
                layer.Update_Weights(alpha)
    def save_weights(self,epoch,patch):
        i=0
        file_name = "Logs/Weights_("+ str(epoch) + ")_(" + str(patch) + ").pkl"
        obj = []
        for l in self.layers:
            if isinstance(l,Layer):
                cache = l.weights
                obj.append(cache)
            i+=1
        with open(file_name, 'wb') as handle:
            pickle.dump(obj, handle, protocol=pickle.HIGHEST_PROTOCOL)

            i+=1
    def load_weights(self,path):
        
        with open(path, 'rb') as f:
            dic = pickle.load(f)

        i=0
        for l in self.layers:
            if isinstance(l,Layer) :
                l.weights['W'] = dic[i]['W'] 
                l.weights['b'] = dic[i]['b']
                i+=1
                


import numpy as np
import cv2 

''' 
    Zero Padding Function 
'''

def ZeroPadding(X ,pad_Width ,dims):
    
    #if dims == (2,3)
    #   pad =[(0,0)(0,0)(P_D,P_D)(P_D,P_D)]
    pad = []
    for idx in range(len(X.shape)):
        if idx not in dims:
            pad.append((0,0))
        else :
            pad.append((pad_Width,pad_Width))
        
    X_padded = np.pad(X, pad, 'constant')
    
    return X_padded
    
    
'''    
pad_width = 2    
dims = (2,3) 
X =   cv2.imread("E:/AUV/DataSet Creation/Final DS/test/2.jpg")
print(X.shape)  
pad = [(0, 0) if idx not in dims else (pad_width, pad_width) for idx in range(len(X.shape))]
X_padded = np.pad(X, pad, 'constant')

print(pad)
print(X_padded)
cv2.imshow("hH",X_padded )
cv2.waitKey(0)
print(pad)
'''


import numpy as np
import PyEZNet.DataLoader as DataLoader

def GetData():
    print('Loadind data......')
    num_classes = 10
    train_images = DataLoader.train_images() #[60000, 28, 28]
    train_images= np.array(train_images,dtype=np.float32)
    train_labels = DataLoader.train_labels()
    test_images = DataLoader.test_images()
    test_images= np.array(test_images,dtype=np.float32)
    test_labels = DataLoader.test_labels()

    print('Preparing data......')
    '''
    Reshaping and normalization training dataset and its labels
    '''
    training_data = train_images.reshape(60000, 1, 28, 28)
    training_data=training_data/255
    training_labels= train_labels.reshape (-1,1)

    '''
    Reshaping and normalization testing dataset and its labels
    '''
    testing_data = test_images.reshape(10000, 1, 28, 28)
    testing_data=testing_data/255
    testing_labels=test_labels .reshape (-1,1)
    
    return training_data,training_labels,testing_data,testing_labels



# Generic Class for the classes that needs to calc Grads

class Diff_Func:
    """
        Abstract Class to represent the Construction of the Differientiable Functions
    """
    def __init__ (self):
    # Init dic for the cached values and grads
        self.cache = {}
        self.grad = {}
        
    def __call__(self, *arg,**kargs):
        # Calc Forward path
        output = self.forward(*arg,**kargs)
        #Calc grads
        self.grad = self.calc_Grad(*arg,**kargs)
        return output
    
    def forward(self , *arg,**kargs):
        """
        Calc output
        """
        pass
    def calc_Grad(self , *arg,**kargs):
        """
        Calc Grads
        """
        pass
    
    def backward(self , *arg,**kargs):
        """
        Backward path
        """
        pass
    



import numpy
import numpy as np
import math
from math import exp
from PyEZNet.Diff_Func import Diff_Func


'''Activation Functions:
       1. Linear or identity
       2. Sinusoid
       3. Tanh
       4. Sigmoid, Logistic, or softstep
       5. Relu
       6. Leaky Relu
       7. Hard Tanh
'''

#Activation Function of linear or identity
class Linear(Diff_Func):
    def forward(self, x):
        return x

    def backward(self, dY):
        return dY * self.grad['x']


    def calc_Grad(self, x):
        y = 1
        grads = {'x' : y}
        return grads

#Activation Function of Sinusoid:
class Sin(Diff_Func):
    def forward(self, x):
        y = sin(x)
        return y
    
    def backward(self, dY):
        return dY * self.grad['x']

    def calc_Grad(self, x):
        y = cos(x)
        grads = {'x' : y}
        return grads

#Activation Function of Tanh
class Tanh(Diff_Func):
    def forward(self, x):
        y = np.tanh(x)
        return y

    def backward(self, dY):
        return dY * self.grad['x']

    def calc_Grad(self, x):
        y = 1.0 - (np.tanh(x))**2
        grads = {'x' : y}
        return grads

#Activation Function of Sigmoid, Logistic, or softstep
class Sigmoid(Diff_Func):
    def forward(self, x):
        y = 1/(1+np.exp(-x))
        return y
    
    def backward(self, dY):
        return dY * self.grad['x']

    def calc_Grad(self, x):
        s=1/(1+np.exp(-x))
        y = s*(1-s)
        grads = {'x' : y}
        return grads

#Activation Function of Relu
class Relu(Diff_Func):
    def forward(self, x):
        return x*(x > 0)
    
    def backward(self, dY):
        return dY * self.grad['X']

    def calc_Grad(self, x):
        
        grads = {'X': 1*(x > 0)}
        return grads

#Activation Function of Leaky Relu
######## alpha=0.01, 3 arguments ???
class Leaky_Relu(Diff_Func):
    def forward(self, x, alpha):
        if x<0.0:
                y = x * alpha
                return y
        else:
             return x

    def backward(self, dY):
        return dY * self.grad['x']

    def calc_Grad(self, x, alpha):
        if x<0.0:
            y = alpha
            grads = {'x' : y}
            return grads 
        else:
            y = 1
            grads = {'x' : y}
            return grads


#Activation Function of Hard Tanh
class Hard_Tanh(Diff_Func):
    def forward(self, x):
        if x <= -1.0:
            return -1
        elif 1.0<x<-1.0:
            return x
        else:
            return 1
        
    def backward(self, dY):
        return dY * self.grad['x']

    def calc_Grad(self, x):
        if x <= -1.0:
            y = 0
            grads = {'x' : y}
            return grads
        elif 1.0<x<-1.0:
            y = 1
            grads = {'x' : y}
            return grads
        else:
            y = 0
            grads = {'x' : y}
            return grads

#Activation Function of Softmax 
class Softmax(Diff_Func):
    
    def forward(self, x):
        out = np.zeros(shape=(32,10))
        x -= np.max(x)
        for i in range(x.shape[0]):
            out[i] = (np.exp(x[i]).T / np.sum(np.exp(x[i]), axis=0)).T
        self.y = out
        return self.y
    

    def backward(self, dY):
        grad = np.zeros(shape=(1,10))
        out = np.zeros(shape=(32,10))
        for i in range(dY.shape[0]):
            s = self.y[i].reshape(-1, 1)
            grad = np.diagflat(s) - np.dot(s, s.T)
            out[i] = np.dot(grad  , dY[i])
            
        return out

    def local_grad(self, x):
        pass
    '''
    def forward(self, x):
        exp_x = np.exp(x)
        y = exp_x / exp_x.sum(axis=0) 
        
        return y

    def backward(self, dY):
        return dY * self.grad['x']


    def local_grad(self, x):
        exp_x = np.exp(x)
        y = exp_x / exp_x.sum(axis=0) 
        # Reshape the 1-d softmax to 2-d so that np.dot will do the matrix multiplication
        s = y.reshape(-1,1)
        gradient=np.diagflat(s) - np.dot(s, s.T)
        grads={'x': gradient}
        return grads
    '''







import numpy as np
from PyEZNet.Diff_Func import Diff_Func


class Loss(Diff_Func):
    """
        Abstract Class to represent the Construction of the Loss Functions
    """
    def forward(self, X, Y):
        """
        Computes the loss of x with respect to y.
        Args:
        X: numpy.ndarray of shape (n_batch, n_dim).
        Y: numpy.ndarray of shape (n_batch, n_dim).
        Returns:
        loss: numpy.float.
        """
        pass
    def backward(self):
        """
        Backward pass for the loss function. Since it should be the final layer
        of an architecture, no input is needed for the backward pass.
        Returns:
        same grad value from calc_Grad()
        """
        return self.grad['X']
    def calc_Grad(self, X, Y):
        """
        Local gradient with respect to X at (X, Y).
        Args:
        X: numpy.ndarray of shape (n_batch, n_dim).
        Y: numpy.ndarray of shape (n_batch, n_dim).
        Returns:
        gradX: numpy.ndarray of shape (n_batch, n_dim).
        """
        pass


class Cross_Entropy_Loss(Loss):


    def forward(self, X, Y,total_loss="mean"):
        #X:is np array with shape (patch,dimen)
        #Y:is np array with shape (patch,1)
        #value of Y in every patch is the truth ground dimension.
        X_probs= Cross_Entropy_Loss.softmax(X)
        loss_array=np.zeros(Y.size)
        self.cache['probs']=X_probs
        for i in range (Y.size):
            #calculating the loss of every patch,prob. of the corresponding ground truth only as all other values will be *(0)
            loss_array[i]= -(np.log(X_probs[i,Y[i]]))
        if(total_loss=="mean"):
            return np.mean(loss_array)
        elif (total_loss=="sum"):
            return np.sum(loss_array)
        elif (total_loss =="median"):
            return np.median(loss_array)




    def calc_Grad(self, X, Y):
        #X:is np array with shape (patch,dimen)
        #Y:is np array with shape (patch,1)
        #value of Y in every patch is the truth ground dimension.
        '''
        probs=self.cache['probs']
        temp_grad= probs
        for i in range (Y.size):
            temp_grad[i,Y[i]]-=1
        grad={}
        grad['X']=temp_grad
        return grad
        '''
        probs = self.cache['probs']
        ones = np.zeros_like(probs)
        for row_idx, col_idx in enumerate(Y):
            ones[row_idx, col_idx] = 1.0

        grads = {'X': (probs - ones)/float(len(X))}
        return grads


    @staticmethod
    def softmax(X):
        X_exp = np.exp(X)
        sum_exp = np.sum(X_exp, axis=1, keepdims=True)
        return (X_exp / sum_exp)



class MeanSquareLoss(Loss):
    """
        Abstract Class to represent the Construction of the Loss Functions
    """
    def forward(self, X, Y):
        """
        X=output of layers
        Y=True labels
        """
        err = np.zeros(X.size)
        squ_sub = np.zeros(X.size)
        err = np.subtract(X,Y)
        squ_sub = np.square(err)
        sum = np.sum(squ_sub,axis=1 ,keepdims= True)
        total_loss = np.average(sum)

        return total_loss



    def calc_Grad(self, X, Y):
        """
        Local gradient with respect to X at (X, Y).
        Args:
        """
        grad = {}
        grad['X'] = 2*(X-Y)

        return grad

class Hinge_Loss(Loss):
    """
        Abstract Class to represent the Construction of the Hinge Loss Function used for binary classification.
    """
    def forward(self, X, Y,total_loss="average"):
        """
        Computes the loss of x with respect to y.
        Args:
        X: numpy.ndarray of shape (n_batch,1).
        Y: numpy.ndarray of shape (n_batch, 1).
        Returns:
        loss: numpy.float.
        """
        zeros_array=np.zeros(X.size)
        ones_array=np.ones(X.size)
        Loss_array=np.zeros(X.size)
        loss_mul=np.zeros(X.size)
        modified_mul=np.zeros(X.size)
        for i in range (X.size):
            loss_mul[i]=((Y[i])*(X[i]))
            modified_mul[i]=ones_array[i]-loss_mul[i]
        Loss_array=np.maximum(zeros_array,modified_mul)
        self.cache['mul']=loss_mul
        if(total_loss=="average"):
            return np.average(Loss_array)
        elif (total_loss=="sum"):
            return np.sum(Loss_array)
        elif (total_loss =="median"):
            return np.median(Loss_array)


    def calc_Grad(self, X, Y):
        """
        Computes the grad dictionary of x with respect to y.
        Args:
        X: numpy.ndarray of shape (n_batch,1).
        Y: numpy.ndarray of shape (n_batch, 1).
        Returns:
        grad: grad dictionary.
        """
        grad_mul=self.cache['mul']
        temp_grad=np.zeros(X.size)
        for i in range (X.size):
            if(grad_mul[i]<1):
                temp_grad[i]=-grad_mul[i]
            else:
                temp_grad[i]=0
        grad={}
        grad['X']=temp_grad
        return grad






import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='PyEZNet',
    version='0.0.1',
    author="Ahmed Khaled",
    author_email="ahmedkhaled11119999@gmail.com",
    description='A neural network framework based on numpy only',
    packages=setuptools.find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/EslamAsfour/PyNNN",
    classifiers=[
        "License :: OSI Approved :: MIT License",
    ],
    install_requires=['numpy'],
    python_requires='>=3.6',
    )
